@@ -1,7 +1,4 @@
 
-from datetime import datetime, time
-
-from django.apps import apps
 from django.utils.functional import cached_property
 from django.urls import reverse_lazy
 from django.core.exceptions import ObjectDoesNotExist
@@ -11,16 +8,14 @@ from django.utils.translation import ugettext_lazy as _
 from django.template.loader import render_to_string
 from django.contrib.auth import get_user_model
 
-from customers.models import CustomerField
-
-from basement.services import register_service
+from invoices.conf import IS_ROUNDING_ENABLED
+from customers.models import CustomerField, Customer
+from suppliers.models import SupplierField, Supplier
+from managers.models import ManagerField, Manager
 
 from exchange.utils import get_price_factory
 from exchange.constants import CURRENCY_UAH
-from exchange.models import MultiCurrencyPrice, ExchangeRates
-
-
-app_config = apps.get_app_config('invoices')
+from exchange.models import MultiCurrencyPrice
 
 
 class InvoiceField(models.ForeignKey):
@@ -60,6 +55,7 @@ class Invoice(models.Model):
 
     type = NotImplemented
     customer = NotImplemented
+    manager = NotImplemented
 
     creator = models.ForeignKey(
         get_user_model(),
@@ -78,6 +74,27 @@ class Invoice(models.Model):
 
         self._old_type = self.type
 
+    @transaction.atomic
+    def update(self, data):
+
+        if 'manager' in data:
+            if data['manager']:
+                self.manager = Manager.objects.get(pk=data['manager'])
+            else:
+                self.manager = None
+
+        if 'customer' in data:
+            if data['customer']:
+                self.customer = Customer.objects.get(pk=data['customer'])
+            else:
+                self.customer = None
+            self.discount = self.customer_discount
+
+        if 'discount' in data:
+            self.discount = int(data['discount'])
+
+        self.save()
+
     def save(self, **kwargs):
         if self.type != self._old_type:
             self.created = timezone.now()
@@ -96,6 +113,12 @@ class Invoice(models.Model):
             return self.customer.name
         return ''
 
+    @property
+    def manager_name(self):
+        if self.manager:
+            return self.manager.name
+        return ''
+
     @property
     def invoice_type(self):
         return self.__class__.__name__.lower()
@@ -109,17 +132,13 @@ class Invoice(models.Model):
         return self._get_url('print')
 
     @property
-    def set_customer_url(self):
-        return self._get_url('set-customer')
+    def update_url(self):
+        return self._get_url('update')
 
     @property
     def add_item_url(self):
         return self._get_url('add-item')
 
-    @property
-    def set_discount_url(self):
-        return self._get_url('set-discount')
-
     def _get_url(self, name):
         return reverse_lazy('invoices:' + name, args=[
             self.invoice_type,
@@ -151,43 +170,10 @@ class Invoice(models.Model):
     def _handle_add_item(self, product, qty, update_product):
         pass
 
-    @transaction.atomic
-    def set_item_qty(self, item_id, value):
-
-        item = self.items.select_related('product').get(pk=item_id)
-
-        if item.qty != value:
-            self._handle_set_item_qty(item, value)
-
-            item.qty = value
-            item.save(update_fields=['qty'])
-
-        return item
-
-    @transaction.atomic
-    def set_item_price(self, item_id, value):
-
-        item = self.items.select_related('product').get(pk=item_id)
-
-        if item.price_uah != value:
-            self._handle_set_item_price(item, value)
-
-            item.initial_currency = CURRENCY_UAH
-            item.price_retail = value
-            item.save()
-
-        return item
-
-    def _handle_set_item_qty(self, item, value):
-        pass
-
-    def _handle_set_item_price(self, item, value):
-        pass
-
     @transaction.atomic
     def remove_item(self, item_id):
 
-        item = self.items.select_related('product').get(pk=item_id)
+        item = self.get_item(item_id)
 
         self._handle_remove_item(item)
 
@@ -200,12 +186,6 @@ class Invoice(models.Model):
     def _handle_remove_item(self, item):
         pass
 
-    def set_customer(self, customer):
-
-        self.customer = customer
-        self.discount = self.customer_discount
-        self.save(update_fields=['customer', 'discount'])
-
     @property
     def total(self):
         return sum([i.subtotal for i in self.items.all()])
@@ -237,10 +217,6 @@ class Invoice(models.Model):
 
         return (self.discount * number) / 100.0
 
-    def set_discount(self, value):
-        self.discount = value
-        self.save(update_fields=['discount'])
-
     def serialize_totals(self):
         return {
             'grand_total': self.total,
@@ -249,9 +225,15 @@ class Invoice(models.Model):
             'total_with_discount': self.total_with_discount
         }
 
+    def get_item(self, item_id):
+        return self.items.select_related('product').get(pk=item_id)
+
     def get_items(self):
         return self.items.all().order_by('-id')
 
+    def get_services(self):
+        return self.services.all().order_by('-id')
+
     class Meta:
         abstract = True
 
@@ -267,6 +249,24 @@ class InvoiceItem(MultiCurrencyPrice):
 
     qty = models.FloatField(_('Quantity'))
 
+    @transaction.atomic
+    def update(self, data):
+
+        if 'qty' in data:
+            self._set_qty(float(data['qty']))
+
+        if 'price' in data:
+            self._set_price(float(data['price']))
+
+        self.save()
+
+    def _set_qty(self, value):
+        self.qty = value
+
+    def _set_price(self, value):
+        self.initial_currency = CURRENCY_UAH
+        self.price_retail = value
+
     def get_qty_input_value(self):
         return str(self.qty).replace(',', '')
 
@@ -286,24 +286,8 @@ class InvoiceItem(MultiCurrencyPrice):
         return self.product.bar_code
 
     @property
-    def set_qty_url(self):
-        return reverse_lazy('invoices:set-item-qty', args=[
-            self.invoice.invoice_type,
-            self.invoice.pk,
-            self.pk
-        ])
-
-    @property
-    def set_price_url(self):
-        return reverse_lazy('invoices:set-item-price', args=[
-            self.invoice.invoice_type,
-            self.invoice.pk,
-            self.pk
-        ])
-
-    @property
-    def remove_url(self):
-        return reverse_lazy('invoices:remove-item', args=[
+    def api_url(self):
+        return reverse_lazy('invoices:item', args=[
             self.invoice.invoice_type,
             self.invoice.pk,
             self.pk
@@ -316,7 +300,7 @@ class InvoiceItem(MultiCurrencyPrice):
     def price(self):
         price = super().price
 
-        if self.currency == CURRENCY_UAH:
+        if IS_ROUNDING_ENABLED and self.currency == CURRENCY_UAH:
             return round(price)
 
         return price
@@ -347,7 +331,7 @@ class InvoiceItem(MultiCurrencyPrice):
 
     @property
     def wholesale_subtotal_uah(self):
-         return self.price_wholesale_uah * self.qty
+        return self.price_wholesale_uah * self.qty
 
     @property
     def profit_uah(self):
@@ -390,18 +374,25 @@ class Arrival(Invoice):
 
     customer = CustomerField(related_name='arrivals')
 
+    supplier = SupplierField(related_name='arrivals')
+
+    manager = ManagerField(related_name='arrivals')
+
+    def update(self, data):
+
+        if 'supplier' in data:
+            if data['supplier']:
+                self.supplier = Supplier.objects.get(pk=data['supplier'])
+            else:
+                self.supplier = None
+
+        super().update(data)
+
     def _handle_add_item(self, product, qty, update_product):
 
         if update_product:
             product.add_stock(value=qty)
 
-    def _handle_set_item_qty(self, item, value):
-
-        if item.qty > value:
-            item.product.subtract_stock(item.qty - value)
-        else:
-            item.product.add_stock(value - item.qty)
-
     def _handle_remove_item(self, item):
 
         if item.qty > 0:
@@ -416,6 +407,20 @@ class ArrivalItem(InvoiceItem):
 
     invoice = InvoiceField(Arrival)
 
+    def _set_qty(self, value):
+
+        if self.qty > value:
+            self.product.subtract_stock(self.qty - value)
+
+        if self.qty < value:
+            self.product.add_stock(value - self.qty)
+
+        super()._set_qty(value)
+
+    def _set_price(self, value):
+        if hasattr(self.product, 'handle_arrival_item_price_change'):
+            self.product.handle_arrival_item_price_change(value)
+
 
 class Sale(Invoice):
 
@@ -424,33 +429,28 @@ class Sale(Invoice):
     TYPE_ONLINE = 3
     TYPE_CUSTOM = 4
     TYPE_DEBT = 5
+    TYPE_IN_PROCESS = 6
 
     TYPES = (
         (TYPE_CASH_REGISTER, _('Cash register')),
         (TYPE_WRITE_OFF, _('Write off')),
         (TYPE_DEBT, _('Debt')),
+        (TYPE_IN_PROCESS, _('In process')),
         (TYPE_CUSTOM, _('Custom')),
+        (TYPE_ONLINE, _('Online')),
     )
 
-    if app_config.is_online_sale_enabled:
-        TYPES += ((TYPE_ONLINE, _('Online')), )
-
     type = InvoiceTypeField(TYPES)
 
     customer = CustomerField(related_name='sales')
 
+    manager = ManagerField(related_name='sales')
+
     def _handle_add_item(self, product, qty, update_product):
 
         if update_product:
             product.subtract_stock(value=qty)
 
-    def _handle_set_item_qty(self, item, value):
-
-        if value > item.qty:
-            item.product.subtract_stock(value - item.qty)
-        else:
-            item.product.add_stock(item.qty - value)
-
     def _handle_remove_item(self, item):
 
         if item.qty > 0:
@@ -460,9 +460,12 @@ class Sale(Invoice):
     def service_total(self):
         return sum([s.subtotal for s in self.services.all()])
 
-    def set_customer(self, customer):
-        super().set_customer(customer)
-        self.services.all().update(customer=customer)
+    def update(self, data):
+
+        super().update(data)
+
+        if 'customer' in data:
+            self.services.all().update(customer=data['customer'])
 
     def serialize_totals(self):
         totals = super().serialize_totals()
@@ -478,203 +481,12 @@ class SaleItem(InvoiceItem):
 
     invoice = InvoiceField(Sale)
 
+    def _set_qty(self, value):
 
-class InvoicesService(object):
+        if value > self.qty:
+            self.product.subtract_stock(value - self.qty)
 
-    def __init__(self, user):
-        self._user = user
-        self._check_access()
+        if value < self.qty:
+            self.product.add_stock(self.qty - value)
 
-    def create_invoice(self, invoice_type, inner_type):
-        model = self._get_invoice_model(invoice_type)
-        return model.objects.create(type=inner_type, creator=self._user)
-
-    def set_customer(self, invoice_type, invoice_id, customer):
-
-        invoice = self.get_invoice(invoice_type, invoice_id)
-        invoice.set_customer(customer)
-
-        return {
-            'total': invoice.serialize_totals()
-        }
-
-    def set_discount(self, invoice_type, invoice_id, value):
-
-        invoice = self.get_invoice(invoice_type, invoice_id)
-        invoice.set_discount(value)
-
-        return {
-            'total': invoice.serialize_totals()
-        }
-
-    def add_item(self, invoice_type, invoice_id, product):
-
-        invoice = self.get_invoice(invoice_type, invoice_id)
-        item = invoice.add_item(product)
-
-        return {
-            'status': 'OK',
-            'html': item.render(),
-            'item_id': item.id,
-            'product': item.serialize_product(),
-            'total': invoice.serialize_totals()
-        }
-
-    def set_item_qty(self, invoice_type, invoice_id, item_id, value):
-
-        invoice = self.get_invoice(invoice_type, invoice_id)
-
-        item = invoice.set_item_qty(item_id, value)
-
-        return {
-            'product': item.serialize_product(),
-            'total': invoice.serialize_totals()
-        }
-
-    def set_item_price(self, invoice_type, invoice_id, item_id, value):
-
-        invoice = self.get_invoice(invoice_type, invoice_id)
-
-        item = invoice.set_item_price(item_id, value)
-
-        return {
-            'product': item.serialize_product(),
-            'total': invoice.serialize_totals()
-        }
-
-    def remove_item(self, invoice_type, invoice_id, item_id):
-
-        invoice = self.get_invoice(invoice_type, invoice_id)
-
-        product = invoice.remove_item(item_id)
-
-        return {
-            'product': product.serialize(),
-            'total': invoice.serialize_totals()
-        }
-
-    def get_invoice(self, invoice_type, invoice_id):
-
-        model = self._get_invoice_model(invoice_type)
-
-        try:
-            return model.objects.get(pk=invoice_id)
-        except ObjectDoesNotExist:
-            raise Exception(_('Invoice not found'))
-
-    def get_sale(self, sale_id):
-        return self.get_invoice('sale', sale_id)
-
-    def get_model_types(self, invoice_type):
-        model = self._get_invoice_model(invoice_type)
-        return model.TYPES
-
-    def get_invoice_items(
-            self,
-            invoice_type,
-            date_from,
-            date_to,
-            category=None,
-            exclude_categories=None):
-
-        model = self._get_invoice_model(invoice_type)
-
-        invoices = model.objects.filter(
-            created__date__range=[date_from, date_to]
-        ).prefetch_related(
-            'items',
-            'items__product'
-        )
-
-        if category is not None:
-            invoices = invoices.filter(type=category)
-
-        if exclude_categories is not None:
-            invoices = invoices.exclude(type__in=exclude_categories)
-
-        invoices = invoices.order_by('created')
-
-        result = []
-        rates = ExchangeRates.objects.get()
-
-        for invoice in invoices:
-            for item in invoice.items.all().select_related('invoice'):
-                item.set_rates(rates)
-                result.append(item)
-
-        return result
-
-    def get_invoice_totals(self, items):
-        return {
-            'qty': sum([i.qty for i in items]),
-            'wholesale_total': sum([i.wholesale_subtotal_uah for i in items]),
-            'retail_total': sum([i.subtotal_with_discount for i in items]),
-            'discounted_retail_total': sum(
-                [i.discounted_subtotal for i in items]),
-            'profit_total': sum([i.profit_subtotal_uah for i in items])
-        }
-
-    def handle_product_change(self, product):
-        self._handle_product_stock_change(product)
-
-    def _handle_product_stock_change(self, product):
-
-        difference = product.stock - product.initial_stock
-
-        if not difference:
-            return
-
-        if difference > 0:
-            self._create_arrival_for_product(product, difference)
-        else:
-            self._create_write_off_for_product(product, abs(difference))
-
-    def _create_arrival_for_product(self, product, qty):
-
-        arrival = self._get_todays_invoice(Arrival, Arrival.TYPE_INCOME)
-
-        arrival.add_item(product, qty=qty, update_product=False)
-
-    def _create_write_off_for_product(self, product, qty):
-
-        sale = self._get_todays_invoice(Sale, Sale.TYPE_WRITE_OFF)
-
-        sale.add_item(product, qty=qty, update_product=False)
-
-    def _get_todays_invoice(self, model, invoice_type):
-
-        today = timezone.now().date()
-        today_min = timezone.make_aware(datetime.combine(today, time.min))
-        today_max = timezone.make_aware(datetime.combine(today, time.max))
-
-        params = {
-            'creator': self._user,
-            'type': invoice_type
-        }
-
-        try:
-            return model.objects.filter(
-                created__range=(today_min, today_max), **params)[0]
-        except IndexError:
-            return model.objects.create(**params)
-
-    def _get_invoice_model(self, invoice_type):
-
-        models = {
-            'sale': Sale,
-            'arrival': Arrival
-        }
-
-        try:
-            return models[invoice_type]
-        except KeyError:
-            raise Exception('Unknown invoice type: {}'.format(invoice_type))
-
-    def _check_access(self):
-        if not self._user.is_staff:
-            raise Exception(_('Access denied'))
-
-
-@register_service('invoices')
-def _construct_service(services, user, **kwargs):
-    return InvoicesService(user)
+        super()._set_qty(value)
